@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2024 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -35,7 +35,6 @@
  *   forward this exception.
  */
 
-#include "common.h"
 #include "prototypes.h"
 
 NOEXPORT int main_unix(int, char*[]);
@@ -55,6 +54,7 @@ int main(int argc, char* argv[]) { /* execution begins here 8-) */
     mallopt(M_MMAP_THRESHOLD, 4096);
 #endif
     tls_init(); /* initialize thread-local storage */
+    crypto_init(); /* initialize libcrypto */
     retval=main_unix(argc, argv);
     main_cleanup();
     return retval;
@@ -99,6 +99,7 @@ NOEXPORT int main_unix(int argc, char* argv[]) {
         signal(SIGCHLD, signal_handler); /* handle dead children */
         signal(SIGHUP, signal_handler); /* configuration reload */
         signal(SIGUSR1, signal_handler); /* log reopen */
+        signal(SIGUSR2, signal_handler); /* connections */
         signal(SIGPIPE, SIG_IGN); /* ignore broken pipe */
         if(signal(SIGTERM, SIG_IGN)!=SIG_IGN)
             signal(SIGTERM, signal_handler); /* fatal */
@@ -106,9 +107,25 @@ NOEXPORT int main_unix(int argc, char* argv[]) {
             signal(SIGQUIT, signal_handler); /* fatal */
         if(signal(SIGINT, SIG_IGN)!=SIG_IGN)
             signal(SIGINT, signal_handler); /* fatal */
+#endif
+#ifdef USE_FORK
+        setpgid(0, 0); /* create a new process group if needed */
 #endif
         daemon_loop();
+#ifdef USE_FORK
+        s_log(LOG_NOTICE, "Terminating service processes");
+        signal(SIGCHLD, SIG_IGN);
+        signal(SIGTERM, SIG_IGN);
+        kill(0, SIGTERM); /* kill the whole process group */
+        while(wait(NULL)!=-1)
+            ;
+        s_log(LOG_NOTICE, "Service processes terminated");
+#endif
+#if !defined(__vms) && !defined(USE_OS2)
+        delete_pid();
+#endif /* standard Unix */
     } else { /* inetd mode */
+        CLI *c;
 #if !defined(__vms) && !defined(USE_OS2)
         close(fd);
 #endif /* standard Unix */
@@ -118,7 +135,11 @@ NOEXPORT int main_unix(int argc, char* argv[]) {
 #endif
         set_nonblock(0, 1); /* stdin */
         set_nonblock(1, 1); /* stdout */
-        client_main(alloc_client_session(&service_options, 0, 1));
+        c=alloc_client_session(&service_options, 0, 1);
+        tls_alloc(c, ui_tls, NULL);
+        service_up_ref(&service_options);
+        client_main(c);
+        client_free(c);
     }
     return 0;
 }
@@ -128,7 +149,7 @@ NOEXPORT void signal_handler(int sig) {
     int saved_errno;
 
     saved_errno=errno;
-    signal_post(sig);
+    signal_post((uint8_t)sig);
     signal(sig, signal_handler);
     errno=saved_errno;
 }
@@ -176,22 +197,18 @@ NOEXPORT int create_pid(void) {
         s_log(LOG_DEBUG, "No pid file being created");
         return 0;
     }
-    if(global_options.pidfile[0]!='/') {
-        /* to prevent creating pid file relative to '/' after daemonize() */
-        s_log(LOG_ERR, "Pid file (%s) must be full path name", global_options.pidfile);
-        return 1;
-    }
-    global_options.dpid=(unsigned long)getpid();
 
-    /* silently remove old pid file */
+    /* silently remove the old pid file */
     unlink(global_options.pidfile);
+
+    /* create a new pid file */
     pf=open(global_options.pidfile, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL, 0644);
     if(pf==-1) {
         s_log(LOG_ERR, "Cannot create pid file %s", global_options.pidfile);
         ioerror("create");
         return 1;
     }
-    pid=str_printf("%lu\n", global_options.dpid);
+    pid=str_printf("%lu\n", (unsigned long)getpid());
     if(write(pf, pid, strlen(pid))<(int)strlen(pid)) {
         s_log(LOG_ERR, "Cannot write pid file %s", global_options.pidfile);
         ioerror("write");
@@ -200,16 +217,18 @@ NOEXPORT int create_pid(void) {
     str_free(pid);
     close(pf);
     s_log(LOG_DEBUG, "Created pid file %s", global_options.pidfile);
-    atexit(delete_pid);
     return 0;
 }
 
 NOEXPORT void delete_pid(void) {
-    if((unsigned long)getpid()!=global_options.dpid)
-        return; /* current process is not main daemon process */
-    s_log(LOG_DEBUG, "removing pid file %s", global_options.pidfile);
-    if(unlink(global_options.pidfile)<0)
-        ioerror(global_options.pidfile); /* not critical */
+    if(global_options.pidfile) {
+        if(unlink(global_options.pidfile)<0)
+            ioerror(global_options.pidfile); /* not critical */
+        else
+            s_log(LOG_DEBUG, "Removed pid file %s", global_options.pidfile);
+    } else {
+        s_log(LOG_DEBUG, "No pid file to remove");
+    }
 }
 
 #endif /* standard Unix */
@@ -252,20 +271,28 @@ void ui_new_log(const char *line) {
 
 /**************************************** ctx callbacks */
 
-int passwd_cb(char *buf, int size, int rwflag, void *userdata) {
-    (void)buf; /* squash the unused parameter warning */
-    (void)size; /* squash the unused parameter warning */
-    (void)rwflag; /* squash the unused parameter warning */
-    (void)userdata; /* squash the unused parameter warning */
-    return 0; /* not implemented */
+int ui_passwd_cb(char *buf, int size, int rwflag, void *userdata) {
+    return PEM_def_callback(buf, size, rwflag, userdata);
 }
 
 #ifndef OPENSSL_NO_ENGINE
-int pin_cb(UI *ui, UI_STRING *uis) {
-    (void)ui; /* squash the unused parameter warning */
-    (void)uis; /* squash the unused parameter warning */
-    return 0; /* not implemented */
+
+int (*ui_get_opener(void)) (UI *) {
+    return UI_method_get_opener(UI_OpenSSL());
 }
+
+int (*ui_get_writer(void)) (UI *, UI_STRING *) {
+    return UI_method_get_writer(UI_OpenSSL());
+}
+
+int (*ui_get_reader(void)) (UI *, UI_STRING *) {
+    return UI_method_get_reader(UI_OpenSSL());
+}
+
+int (*ui_get_closer(void)) (UI *) {
+    return UI_method_get_closer(UI_OpenSSL());
+}
+
 #endif
 
 /* end of ui_unix.c */
